{-
	Denotational Semantics
	A CCS-based language

	First approach with own extensions!
-}

infix `equalTr`, `equalNat`, `plusNat`, `timesNat`

infixr 0 !!!
infixr 1 &&&, &&>

------------------------------------------------------------------------------
--  Abstract Syntax:  --------------------------------------------------------
------------------------------------------------------------------------------

{-

P of Program
C of Command
E of Expression

I of Identifier
N of Numeral

P ::= C '.'
C ::= C1 ';' C2 | C1 '||' C2 | I := E | 'if' E 'then' C1 'else' C2 |
      'while' E 'do' C | 'skip'

E ::= E1 '=' E2 | '~' E | 'true' | 'false' |	(Boolean expression)
      E1 '+' E2 | E1 '*' E2 | '(' E ')' | I | N	(plain expression)

-}

data Program =
	  Dot Command

data Command =
	  Seq Command Command
	| Par Command Command
	| Assign Identifier Expression
	| If Expression Command Command
	| While Expression Command
	| Skip

data Expression =
	  Equal Expression Expression
	| Not Expression
	| TrueVal
	| FalseVal
	| Plus Expression Expression
	| Times Expression Expression
	| Braced Expression
	| Id Identifier
	| Num Numeral

data Identifier =
	  Ident String

data Numeral =
	  Number String

------------------------------------------------------------------------------
-- Syntax analysis -----------------------------------------------------------
------------------------------------------------------------------------------

data Terminal = 
	  TDot
	| TSemicolon
	| TPar
	| TAssign
	| TIf
	| TThen
	| TElse
	| TWhile
	| TDo
	| TSkip
	| TEqual
	| TNot
	| TTrue
	| TFalse
	| TPlus
	| TTimes
	| TLpar
	| TRpar
	| TName
	| TNumber
	| TEOF

instance Eq Terminal where
	TDot == TDot = True
	TSemicolon == TSemicolon = True
	TPar == TPar = True
	TAssign == TAssign = True
	TIf == TIf = True
	TThen == TThen = True
	TElse == TElse = True
	TWhile == TWhile = True
	TDo == TDo = True
	TSkip == TSkip = True
	TEqual == TEqual = True
	TNot == TNot = True
	TTrue == TTrue = True
	TFalse == TFalse = True
	TPlus == TPlus = True
	TTimes == TTimes = True
	TLpar == TLpar = True
	TRpar == TRpar = True
	TName == TName = True
	TNumber == TNumber = True
	_ == _ = False

---- Scanner -----------------------------------------------------------------

getNextToken :: String -> ( Terminal, String, String )

getNextToken [] = ( TEOF, "", [] )
getNextToken ('.':cs) = ( TDot, "", cs )
getNextToken (';':cs) = ( TSemicolon, "", cs )
getNextToken ('|':'|':cs) = ( TPar, "", cs )
getNextToken (':':'=':cs) = ( TAssign, "", cs )

getNextToken ('=':cs) = ( TEqual, "", cs )
getNextToken ('~':cs) = ( TNot, "", cs )

getNextToken ('+':cs) = ( TPlus, "", cs )
getNextToken ('*':cs) = ( TTimes, "", cs )
getNextToken ('(':cs) = ( TLpar, "", cs )
getNextToken (')':cs) = ( TRpar, "", cs )

getNextToken (c:cs) | isSpace c = getNextToken cs
                    | isDigit c = ( TNumber, (c:ns), cs' )
                    | isAlpha c = ( (getKey (c:ns')), (c:ns'), cs'' )
                    | otherwise = getNextToken cs
                    where (ns, cs' )  = ( getWhile isDigit cs ,
                                          getRemain isDigit cs )
                          (ns', cs'') = ( getWhile isAlphanum cs,
                                          getRemain isAlphanum cs )

getKey :: String -> Terminal

getKey "if" = TIf
getKey "then" = TThen
getKey "else" = TElse
getKey "while" = TWhile
getKey "do" = TDo
getKey "skip" = TSkip
getKey "true" = TTrue
getKey "false" = TFalse
getKey _ = TName 

getWhile :: (Char -> Bool) -> String -> String
getWhile _ [] = []
getWhile p (c:cs) = if p c
                    then c : getWhile p cs
                    else []

getRemain :: (Char -> Bool) -> String -> String
getRemain _ [] = []
getRemain p css@(c:cs) = if p c
                         then getRemain p cs
                         else css

---- Parser -------------------------------------------------------------------

type ParserState = String

data ParserResult a =
	  Succ a ParserState
	| Fail

type Parser a = ParserState -> ParserResult a


epsilon :: a -> Parser a
epsilon a = \ps -> Succ a ps


literal :: Terminal -> Parser ()
literal t = \s -> let (nt, v, sr) = getNextToken s in
                            if nt == t
                            then Succ () sr
                            else Fail 

value :: Terminal -> Parser String
value t = \s -> let (nt, v, sr) = getNextToken s in
                            if nt == t
                            then Succ v sr
                            else Fail

(!!!) :: Parser a -> Parser a -> Parser a

p1 !!! p2 = \s -> case p1 s of
			result@(Succ _ _) -> result
			Fail -> p2 s

(&&&) :: Parser a -> Parser b -> Parser b

p1 &&& p2 = \s -> case p1 s of
			Succ v1 s1 -> case p2 s1 of
					result@(Succ _ _) -> result
					Fail -> Fail
			Fail -> Fail

(&&>) :: Parser a -> (a -> Parser b) -> Parser b

p1 &&> p2 = \s -> case p1 s of
			Succ v1 s1 -> case p2 v1 s1 of
					result@(Succ _ _) -> result
					Fail -> Fail
			Fail -> Fail

---- language parser --------------------------------------------------------

parse_P :: Parser Program

parse_P =
	parse_C &&> (\c -> literal TDot &&& epsilon (Dot c) )

parse_C :: Parser Command

parse_C =
	parse_C' &&> (\c1 -> literal TSemicolon &&& parse_C &&>
			(\c2 -> epsilon (Seq c1 c2) ) )

	!!!

	parse_C' &&> (\c1 -> literal TPar &&& parse_C &&>
			(\c2 -> epsilon (Par c1 c2) ) )

	!!!

	parse_C'

parse_C' :: Parser Command

parse_C' =
	parse_I &&> (\id -> literal TAssign &&& parse_E &&>
			(\e -> epsilon (Assign id e) ) )

	!!!

	literal TIf &&& parse_E &&>
		(\e -> literal TThen &&& parse_C &&>
		  (\c1 -> literal TElse &&& parse_C &&>
		    (\c2 -> epsilon (If e c1 c2) ) ) )

	!!!

	literal TWhile &&& parse_E &&>
		(\e -> literal TDo &&& parse_C &&>
		  (\c -> epsilon (While e c) ) )

	!!!

	literal TSkip &&& epsilon (Skip)


parse_E :: Parser Expression

parse_E =
	parse_E' &&> (\e1 -> literal TEqual &&& parse_E &&>
			(\e2 -> epsilon (Equal e1 e2) ) )

	!!!

	parse_E' &&> (\e1 -> literal TPlus &&& parse_E &&>
			(\e2 -> epsilon (Plus e1 e2) ) )

	!!!

	parse_E' &&> (\e1 -> literal TTimes &&& parse_E &&>
			(\e2 -> epsilon (Times e1 e2) ) )

	!!!

	parse_E'


parse_E' :: Parser Expression

parse_E' =
	literal TNot &&& parse_E &&> (\e -> epsilon (Not e) )

	!!!

	literal TTrue &&& epsilon (TrueVal)

	!!!

	literal TFalse &&& epsilon (FalseVal)

	!!!

	literal TLpar &&& parse_E &&>
		(\e -> literal TRpar &&& epsilon (Braced e) )

	!!!

	parse_I &&> (\id -> epsilon (Id id) )

	!!!

	parse_N &&> (\num -> epsilon (Num num) )

parse_I :: Parser Identifier

parse_I =
	value TName &&> (\i -> epsilon (Ident i) )

parse_N :: Parser Numeral

parse_N =
	value TNumber &&> (\n -> epsilon (Number n) )

-- interpreter ---------------------------------------------------------------

interpreter = \s -> parse_P s

------------------------------------------------------------------------------
--  Semantic algebras:  ------------------------------------------------------
------------------------------------------------------------------------------

{-

I.	Truth values
	Domain t of Tr = Boolean
	Operations
	  true, false : Tr
	  notTr : Tr -> Tr
	  equalTr : Tr x Tr -> Tr

-}

type Tr = Bool

true = True
false = False

notTr :: Tr -> Tr
notTr t = not t

equalTr :: Tr -> Tr -> Tr
b1 `equalTr` b2 = b1 == b2

------------------------------------------------------------------------------

{-

II.	Numerals
	Domain n of Nat = Int
	Operations
	  0, 1, 2 ... : Nat
	  equalNat : Nat x Nat -> Tr
	  plusNat : Nat x Nat -> Nat
	  timesNat : Nat x Nat -> Nat
-}

type Nat = Int

equalNat :: Nat -> Nat -> Tr

n1 `equalNat` n2 = n1 == n2

plusNat :: Nat -> Nat -> Nat

n1 `plusNat` n2 = n1 + n2

timesNat :: Nat -> Nat -> Nat

n1 `timesNat` n2 = n1 * n2

------------------------------------------------------------------------------

{-

III.	Identifiers
	Domain i of Id = Identifier

	(no language representation needed)
-}

------------------------------------------------------------------------------

{-

IV.	Storable-Values (to denote values which may be stored)
	Domain v of StorableValue = Tr + Nat

-}

data StorableValue =
	  TrVal Tr
	| NatVal Nat

------------------------------------------------------------------------------

{-

V.	Expressible-Values (to denote expressions)
	Domain x of ExpressibleValue = StorableValue + ErrValue
-}

data ExpressibleValue =
	  Expr StorableValue
	| ErrValue

isErrValue :: ExpressibleValue -> Bool
isErrValue ErrValue = True
isErrValue _ = False

isStorableValue :: ExpressibleValue -> Bool
isStorableValue v = not (isErrValue v)

inErrValue :: ExpressibleValue
inErrValue = ErrValue

------------------------------------------------------------------------------

{-

VI.	Store
	Domain s of Store = [Id x StorableValue]
	(Id is handled as String)
	Operations
	  newstore : Store
	  access : Id -> Store -> StorableValue
	  update : Id -> StorableValue -> Store -> Store
-}

type Store = [(String, StorableValue)]

newstore :: Store
newstore = []

access :: String -> Store -> StorableValue
access id [] = (NatVal 0) -- an access to a not stored value delivers Nat 0
access id ((i,v):sr) | id == i = v
		     | otherwise = access id sr

update :: String -> StorableValue -> Store -> Store
update id v [] = [(id,v)]
update id v ((i,old):sr) | id == i = (i,v):sr
		         | otherwise = (i,old) : update id v sr

------------------------------------------------------------------------------

{-

VII.	PostState

-}

------------------------------------------------------------------------------

{-

VIII.	Store manager behavior expression

	sem(s) = _alpha s.gamma s'. sem(s')
-}

------------------------------------------------------------------------------



