{-
	Denotational Semantics
	A calculator language

Denotational Definition:

Abstract Syntax:

P of Program
S of Expr-sequence
E of Expression
N of Numeral

P ::= 'ON' S
S ::= E 'TOTAL' S | E 'TOTAL' 'OFF'
E ::= E1 '+' E2 | E1 '*' E2 | 'IF' E1 ',' E2 ',' E3 | 
      'LASTANSWER' | '(' E ')' | N

Semantic algebras:

I.	Truth values
	Domain t of Tr = Boolean
	Operations
	 true, false : Tr

II.	Natural numbers
	Domain n of Nat
	Operations
	 zero, one, two, ... : Nat
	 plus, times : Nat x Nat -> Nat
	 equals : Nat x Nat -> Tr

Valuation functions:

PP : Program -> Nat*

  PP['ON' S] = SS[S](zero)

SS : Expr-sequence -> Nat -> Nat*

  SS[E 'TOTAL' S](n) = let n' = EE[E](n) in n' cons SS[S](n')
  SS[E 'TOTAL' 'OFF'](n) = EE[E](n) cons nil

EE : Expression -> Nat -> Nat

  EE[E1 '+' E2](n) = EE[E1](n) plus EE[E2](n)
  EE[E1 '*' E2](n) = EE[E1](n) times EE[E2](n)
  EE['IF' E1 ',' E2 ',' E3](n) = 
           EE[E1](n) equals zero -> EE[[E2]](n) [] EE[E3](n)
  EE['LASTANSWER'](n) = n
  EE['(' E ')'](n) = EE[E]](n)
  EE[N](n) = NN[N]

NN : Numeral -> Nat (omitted - maps numeral N to correspondong n of Nat)

-}

infix `equals`, `plus`, `times`


infixr 0 !!!
infixr 1 &&&, &&>

------------------------------------------------------------------------------
---- Definitions for semantic domains ----------------------------------------
------------------------------------------------------------------------------

---- Truth values ------------------------------------------------------------

type Tr = Bool

{- true and false are predefined but written  True and False -}

---- Natural numbers ---------------------------------------------------------

type Nat = Int

plus :: Nat -> Nat -> Nat
n1 `plus` n2 = n1 + n2

times :: Nat -> Nat -> Nat
n1 `times` n2 = n1 * n2


equals :: Nat -> Nat -> Tr
n1 `equals` n2 | n1 == n2 = True
               | otherwise = False

------------------------------------------------------------------------------
---- Definitions for abstract syntax -----------------------------------------
------------------------------------------------------------------------------

{-
P of Program
S of Expr-sequence
E of Expression
N of Numeral

P ::= 'ON' S
S ::= E 'TOTAL' S | E 'TOTAL' 'OFF'
E ::= E1 '+' E2 | E1 '*' E2 | 'IF' E1 ',' E2 ',' E3 | 
      'LASTANSWER' | '(' E ')' | N
-}


data Numeral = 
	  Num String

data Expression =
	  Plus Expression Expression
	| Times Expression Expression
	| If Expression Expression Expression
	| LastAnswer
	| Braced Expression
	| N Numeral

data ExprSequence =
	  Total Expression ExprSequence
	| Off Expression

data Program =
	  On ExprSequence

------------------------------------------------------------------------------
---- Definitions for valuation functions -------------------------------------
------------------------------------------------------------------------------

pp :: Program -> [Nat]
pp (On s) = ss s 0

ss :: ExprSequence -> Nat -> [Nat]
ss (Total e s) n = let n' = (ee e n) in n' : (ss s n') 
ss (Off e) n = (ee e n) : []

ee :: Expression -> Nat -> Nat
ee (Plus e1 e2) n  = (ee e1 n) `plus` (ee e2 n)
ee (Times e1 e2) n = (ee e1 n) `times` (ee e2 n)
ee (If e1 e2 e3) n | (ee e1 n) `equals` 0 = (ee e2 n)
                   | otherwise = (ee e3 n)
ee (LastAnswer) n  = n
ee (Braced e) n    = (ee e n)
ee (N num) n       = (nn num)


nn :: Numeral -> Nat
nn (Num num) = str2nat num


str2nat :: String -> Nat

str2nat  s = foldl (\s c -> s * 10 + (ord c - ord '0')) 0 s

------------------------------------------------------------------------------
---- Definition for syntax analysis ------------------------------------------
------------------------------------------------------------------------------

{-
P ::= 'ON' S
S ::= E 'TOTAL' S | E 'TOTAL' 'OFF'
E ::= E1 '+' E2 | E1 '*' E2 | 'IF' E1 ',' E2 ',' E3 | 
      'LASTANSWER' | '(' E ')' | N
-}

data Terminal =
	  TOn
	| TTotal
	| TOff
	| TPlus
	| TTimes
	| TIf
	| TComma
	| TLastAnswer
	| TLpar
	| TRpar
	| TNumber
	| TError
	| TEOF

instance Eq Terminal where
    TOn == TOn = True
    TTotal == TTotal = True
    TOff == TOff = True
    TPlus == TPlus = True
    TTimes == TTimes = True
    TIf == TIf = True
    TComma == TComma = True
    TLastAnswer == TLastAnswer = True
    TLpar == TLpar = True
    TRpar == TRpar = True
    TNumber == TNumber = True
    _ == _ = False

---- Scanner -----------------------------------------------------------------

getNextToken :: String -> ( Terminal, String, String )

getNextToken [] = ( TEOF, "", [] )
getNextToken ('+':cs) = ( TPlus, "", cs )
getNextToken ('*':cs) = ( TTimes, "", cs )
getNextToken (',':cs) = ( TComma, "", cs )
getNextToken ('(':cs) = ( TLpar, "", cs )
getNextToken (')':cs) = ( TRpar, "", cs )

getNextToken (c:cs) | isSpace c = getNextToken cs
                    | isDigit c = ( TNumber, (c:ns), cs' )
                    | isAlpha c = ( (getKey (c:ns')), (c:ns'), cs'' )
                    | otherwise = getNextToken cs
                    where (ns, cs' )  = ( getWhile isDigit cs ,
                                          getRemain isDigit cs )
                          (ns', cs'') = ( getWhile isAlphanum cs,
                                          getRemain isAlphanum cs )

getKey :: String -> Terminal

getKey "ON" = TOn
getKey "TOTAL" = TTotal
getKey "OFF" = TOff
getKey "IF" = TIf
getKey "LASTANSWER" = TLastAnswer
getKey _ = TError 

getWhile :: (Char -> Bool) -> String -> String
getWhile _ [] = []
getWhile p (c:cs) = if p c
                    then c : getWhile p cs
                    else []

getRemain :: (Char -> Bool) -> String -> String
getRemain _ [] = []
getRemain p css@(c:cs) = if p c
                         then getRemain p cs
                         else css


scan :: String -> [(Terminal, String, String)]

scan "" = [(TEOF, "", "")]
scan s  = (t, v, sr):(scan sr)
        where (t, v, sr) = getNextToken s

---- Parser -------------------------------------------------------------------

type ParserState = String

data ParserResult a =
	  Succ a ParserState
	| Fail

type Parser a = ParserState -> ParserResult a


epsilon :: a -> Parser a
epsilon a = \ps -> Succ a ps


literal :: Terminal -> Parser ()
literal t = \s -> let (nt, v, sr) = getNextToken s in
                            if nt == t
                            then Succ () sr
                            else Fail 

value :: Terminal -> Parser String
value t = \s -> let (nt, v, sr) = getNextToken s in
                            if nt == t
                            then Succ v sr
                            else Fail

(!!!) :: Parser a -> Parser a -> Parser a

p1 !!! p2 = \s -> case p1 s of
			result@(Succ _ _) -> result
			Fail -> p2 s

(&&&) :: Parser a -> Parser b -> Parser b

p1 &&& p2 = \s -> case p1 s of
			Succ v1 s1 -> case p2 s1 of
					result@(Succ _ _) -> result
					Fail -> Fail
			Fail -> Fail

(&&>) :: Parser a -> (a -> Parser b) -> Parser b

p1 &&> p2 = \s -> case p1 s of
			Succ v1 s1 -> case p2 v1 s1 of
					result@(Succ _ _) -> result
					Fail -> Fail
			Fail -> Fail

---- Calculator parsers ------------------------------------------------------

{-
P ::= 'ON' S
S ::= E 'TOTAL' S | E 'TOTAL' 'OFF'
E ::= E1 '+' E2 | E1 '*' E2 | 'IF' E1 ',' E2 ',' E3 | 
      'LASTANSWER' | '(' E ')' | N
-}

parse_N :: Parser Numeral

parse_N =
	value TNumber &&> (\n -> epsilon (Num n) )


parse_E :: Parser Expression

{-

After transformation:: left-recursive rules --> right-recursive rules
                       epsilon-rules --> epsilon free grammar

E  ::= E + E | E * E | IF E , E , E | LASTANSWER | ( E ) | N

stage 1: left-recursive -> indirect right-recursive

E  ::= IF E , E , E E' | LASTANSWER E' | ( E ) E' | N E'
E' ::= + E | * E | <epsilon>

stage 2: epsilon -> epsilon free

E  ::= E' + E | E' * E | E'
E' ::= IF E , E , E | LASTANSWER | ( E ) | N

-}

parse_E =
	parse_E' &&> (\e1 -> literal TPlus &&& parse_E &&>
			(\e2 -> epsilon (Plus e1 e2) ) )

	!!!

	parse_E' &&> (\e1 -> literal TTimes &&& parse_E &&>
			(\e2 -> epsilon (Times e1 e2) ) )

	!!!

	parse_E'

parse_E' :: Parser Expression

parse_E' =
	literal TIf &&& parse_E &&>
		(\e1 -> literal TComma &&& parse_E &&>
		   (\e2 -> literal TComma &&& parse_E &&>
			(\e3 -> epsilon (If e1 e2 e3) ) ) )

	!!!

	literal TLastAnswer &&& epsilon (LastAnswer)

	!!!

	literal TLpar &&& parse_E &&>
		(\e -> literal TRpar &&& epsilon (Braced e) )

	!!!

	parse_N &&> (\num -> epsilon (N num) )


parse_S :: Parser ExprSequence

parse_S =
	parse_E &&> (\e -> literal TTotal &&& parse_S &&>
			(\s -> epsilon (Total e s) ) )

	!!!

	parse_E &&> (\e -> literal TTotal &&& literal TOff &&&
			epsilon (Off e) )


parse_P :: Parser Program

parse_P =
	literal TOn &&& parse_S &&> (\s -> epsilon (On s) )

------------------------------------------------------------------------------
-- Calculator ----------------------------------------------------------------
------------------------------------------------------------------------------

calculator :: String -> [Nat]

calculator = \s -> case parse_P s of
                         Succ p _ -> pp p
                         Fail -> []

